/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770   */
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566   */
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/*                                                                              */
/* 9010LinuxUartDriver.c - UART driver for Hytec serial IPs for Linux           */
/*                                                                              */
/* (C)2011 Hytec Electronics Ltd.                                               */
/* Developer: Jim S Chen                                                        */
/* Start: 11/01/2011                                                            */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Revision history: (comment and initial revisions)                           */
/*                                                                              */
/*  vers.       revised         modified by                                     */
/*  -----       -----------     -------------------------                       */
/*  1.0			11-JAN-2011		J. Chen, Hytec Electronics Ltd.					*/
/*                              Initial version.                                */
/*  1.1			13-APR-2011		J. Chen, Hytec Electronics Ltd.					*/
/*                              a). Fixed half duplex echoing issue.            */
/*                              b). Added reset individual port at the start    */
/*                              	of setup routine   							*/
/*  1.2			08-JUN-2011		J. Chen, Hytec Electronics Ltd.					*/
/*                              a). Fixed interrupt conflict with keyboard      */
/*                              b). Fixed disabling interrupt in the port       */
/*                                  shut down routine					        */
/*                              											    */
/********************************************************************************/
/*  Notes to this driver: 							                            */
/*             																	*/
/*  1. This driver enables interrupt for all ports of all serial IP cards in 	*/
/*  the system. The interrupt vectors are hard coded starting from 0xC0 and 	*/
/*  onwards. The maximum number of IP cards of a system (in IOC9010 case ) 		*/
/*	could have up to 6 IPs so that their vectors could use 0xC0 ~ 0xC5. So		*/
/*  caution needs to be taken into account when defining interrupt vectors for	*/
/*	other devices such that they don't conflict.                                */
/*                                                                              */
/*  2. To co-exist with Hytec PCI device driver which mapped to the same PCI    */
/*  spaces and resources, we need to somehow tell the PCI driver not to handle  */
/*  interrupts generated by the uart ports. This is implemented by setting a	*/
/*  flag 0xAA in the Timer Control Register LS (0x06) when this driver is 		*/
/*	loaded so that the PCI interrupt service routine can pass the interrupts 	*/
/*	over upon detecting this flag. Of course when this driver is unloaded and 	*/
/*	destroyed, the flag is cleared.							    				*/
/*                                                                              */
/*  This also means that we cannot use the timers in the uart chip. The  		*/
/*  fact is that we are not aware of anybody uses it!                           */
/*                                                                              */
/********************************************************************************/

/**********************************************************************/
/* 					 Includes and Macro Definitions	 				  */
/**********************************************************************/
/*** Generic Includes ***/
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <linux/version.h>
#include "linux/irq.h"
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/errno.h>

/*** serial uart related ***/
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/platform_device.h>

/*** legacy issue ***/
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
#include <linux/config.h>
#else
#include <linux/autoconf.h>
#endif

/*** Hytec specific ***/
#include "9010LinuxUartDriver.h"
#include "Hy85156.h"

//****** version info ***********************/
#define DRIVER_VERSION "v1.0"
#define DRIVER_AUTHOR "Jim S Chen <jim.chen@hytec-electronics.co.uk>"
#define DRIVER_DESC "IOC9010 Serial UART Driver"

/* Module information */
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

#define USHORT unsigned short int
#define ULONG unsigned long int
#define UCHAR unsigned char

#define MAX_UART_NUM 	 		48		/* maximum serial devices defined by probe */
#define IOC9010_UART_MAJOR 		80		/* dynamic major number? tty doesn't like dynamic. so fix to 80 */
int IOC9010_UART_MINOR = 		48;		/* maximum serial devices defined by probe */
dev_t dtMajorMinor;

#define PORT_IOC9010	 		9010
#define FIFO_SIZE		 		64
/**********************************************************************/
/*  						 PCI Device Definitions	 				  */
/**********************************************************************/
static struct pci_device_id ids[] = {
	{ PCI_DEVICE(0x1196, 0x9010), },
	{ PCI_DEVICE(0x1196, 0x6335), },
	{ PCI_DEVICE(0x1196, 0x7002), },
	{ 0, } 
};

/*****Interrupt operation data area **********/
#define PLXLOCALCONFIGCSR		0x4c  	/* Control register   */
#define HYTEC7002CSR			0x00	/* Control register   */
#define PLXLOCALCONFIGCSR8311	0x68  	/* Control register   */
#define IACK					0xa		/* Interrupt acknowledge register   */
#define CSR		    			0xc		/* Control   and Status  register   */

struct IOC9010_dev {
	struct IOC9010_dev* next;			/* store next device pointer */
	int					carrierNum;		/* deice number */
	
	void 				*virtureMemBaseAddr;
	unsigned long 		carrierAddr;  	/* Base Address of Carrier Card Registers - I/O Region */
	unsigned long 		iocMemoryAddr;  /* IOC memory Base for IP cards - memory Region */
	unsigned long 		plxLocalConfigBAR;	/* Base Address of PLX local configuration Registers - I/O Region. */
	unsigned long 		pciExpressBAR;  /* Base Address of PCI express configuration Registers (BAR0)- I/O Region. */

	unsigned char 		carrierIRQ;   	/* Interrupt Number of the Carrier Card   - IRQ */
	unsigned short 		vector;			/* for interrupt. added by Jim on 26-03-2009 */
	
	u16 				ipCard[6];      /* Array of Industry Pack Cards Found */
	int					ipSlotNum;		/* store how many IP slot numbers for this carrier */
	int 				vendor_ID;		/* vendor ID */
	int 				device_ID;		/* device ID */
	int 				pci_bus;		/* pci bus number */
	int 				pci_dev;		/* pci device number */
	int 				pci_func;		/* pci function number */
	
	int 				bar1State;		/* for BAR1 IO/memory allocation successful flag */
	int 				bar2State;		/* for BAR2 IO/memory allocation successful flag */
	int 				bar3State;		/* for BAR3 IO/memory allocation successful flag */

	struct pci_dev 		*ppciDev;		/* store the associated pci device struct pointer */

	/*struct semaphore sem;*/       	/* mutual exclusion semaphore */
  	spinlock_t 			device_lock;   	/* access to device registers */
	unsigned long 		lock_flags;
};

struct IOC9010_dev* pciDevList;			/* the pointer to the pci devices */


/**********************************************************************/
/*  						 UART Device Definitions	 			  */
/**********************************************************************/
struct ioc9010_uart_port {
	struct 				uart_port port;	/* This has to be the first element since it shares the address with its parent */
	int					open_count;		/* number of times this port has been opened */
	struct semaphore	sem;			/* locks this structure */
	struct timer_list	timer;

	/* PCI device base addresses */
	int 				device_ID;		/* device ID */
	void 				*virtureMemBaseAddr;
	unsigned long 		carrierAddr;  	/* Base Address of Carrier Card Registers - I/O Region */
	unsigned long 		plxLocalConfigBAR;	/* Base Address of PLX local configuration Registers - I/O Region. */

	int 				intLevel;
	int 				vector;
	int 				irq;
	int					mask;			/* interrupt mask */
	int					rcvTrigLevel;	/* receive trigger level */
	int					transTrigLevel;	/* transmit trigger level */

	/* serial write/read functions settings */
	unsigned int		old_status;
	int					ipSlot;
	int					portNo;			/* port number */
	int					moduleType;		/* either 8515 or 8516 */
	unsigned long 		baud;			/* baud rate */
	char				parity;			/* parity check */
	int					stopbits;		/* stop bits */
	int 				databits;		/* data bits */
	int					flowControl;	/* flow control flag */
	int					duplex;			/* duplex flag, only for 8516 */
	int					withEcho;		/* need echo flag, only for 8516 */
	int					delay485;		/* delay for RS485 for 8516 */
	int					flag;			/* first time flag */
	
	struct IOC9010_dev* pcidev;			/* parent device pointer */
};

static struct ioc9010_uart_port ioc9010_ports[MAX_UART_NUM];


/**********************************************************************/
/*  						 Internal Functions			 			  */
/**********************************************************************/

/* pci device related routines */
int buildHytecPciDeviceList(void);
int checkHytecPciDevice(struct pci_dev *pPciDev);
int addPciDevice(struct pci_dev *pPciDev);
int removePciDevice(void);

/* register/memory access routines */
long IOC9010IPRead(void *memoryBaseAddress, int ipslot, USHORT add, USHORT *data);
long IOC9010IPWrite(void *memoryBaseAddress, int ipslot, USHORT add, USHORT data);
long IOC9010IPMemoryRead(void *memoryBaseAddress, int ipslot, USHORT add, int len, USHORT *data);
long IOC9010IPMemoryWrite(void *memoryBaseAddress, int ipslot, USHORT add, int len, USHORT *data);

/* serial related routines */
long Hy85156_Configure(struct ioc9010_uart_port *ioc9010, char vector, int delay);
long Hy85156_ComsSetup(struct ioc9010_uart_port *ioc9010, int portNo, ULONG baud, char parity, int numstopbits, int numdatabits, char flowctr);
long Hy8516_ComsSetup(struct ioc9010_uart_port *ioc9010, int portNo, ULONG baud, char parity, int numstopbits, int numdatabits, char flowctr, int halfduplex, int delay485);
long Hy85156_EnableInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int mask, int rcvTrigLevel, int transTrigLevel);
long Hy85156_DisableInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int mask);
long Hy85156_INTReason(struct ioc9010_uart_port *ioc9010, int* portNo, int *reason);
long Hy85156_ResetInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int reason);
long Hy85156_WriteLine(struct ioc9010_uart_port *ioc9010, int portNo, int line, int data);
long Hy85156_Write(struct ioc9010_uart_port *ioc9010, int portNo, const UCHAR *value, int length);
long Hy85156_Read(struct ioc9010_uart_port *ioc9010, int portNo, UCHAR *value);
long Hy85156_Flush_Single(struct ioc9010_uart_port *ioc9010, int portNo, int TX, int RX);
long Hy85156_Flush(struct ioc9010_uart_port *ioc9010, USHORT mask);
long EnableInterrupt(struct ioc9010_uart_port *ioc9010, int level, int side);
long DisableInterrupt(struct ioc9010_uart_port *ioc9010, int level, int side);
static void ioc9010_tx_chars(struct ioc9010_uart_port *ioc9010);
static void ioc9010_rx_chars(struct ioc9010_uart_port *ioc9010);


/********************************************End of Definition************************************************/

/**********************************************************************/
/*  						 Main Business				 			  */
/**********************************************************************/

/*
 * Handle any change of modem status signal since we were last called.
 */
static void ioc9010_mctrl_check(struct ioc9010_uart_port *ioc9010)
{
	unsigned int status, changed;

	status = ioc9010->port.ops->get_mctrl(&ioc9010->port);
	changed = status ^ ioc9010->old_status;

	if (changed == 0)
		return;

	ioc9010->old_status = status;

	if (changed & TIOCM_RI)
		ioc9010->port.icount.rng++;
	if (changed & TIOCM_DSR)
		ioc9010->port.icount.dsr++;
	if (changed & TIOCM_CAR)
		uart_handle_dcd_change(&ioc9010->port, status & TIOCM_CAR);
	if (changed & TIOCM_CTS)
		uart_handle_cts_change(&ioc9010->port, status & TIOCM_CTS);

	wake_up_interruptible(&ioc9010->port.info->delta_msr_wait);
/*	wake_up_interruptible(&ioc9010->port.state->port.delta_msr_wait);kernel2.6.37*/
}

/*
 * This is our per-port timeout handler, for checking the
 * modem status signals.
 */
static void ioc9010_timeout(unsigned long data)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)data;
	unsigned long flags;

/*	if (ioc9010->port.state) {  kernel2.6.37*/
	if (ioc9010->port.info) 
	{
		spin_lock_irqsave(&ioc9010->port.lock, flags);
		ioc9010_mctrl_check(ioc9010);
		spin_unlock_irqrestore(&ioc9010->port.lock, flags);
/*		mod_timer(&ioc9010->timer, jiffies + MCTRL_TIMEOUT);*/
		mod_timer(&ioc9010->timer, jiffies + 25);
	}
}


irqreturn_t ioc9010_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)dev_id;
	int reason, port;
	int INTSource, i=0, INTFound;
	USHORT data;

	spin_lock(&ioc9010->port.lock);

    INTSource = inw(ioc9010->carrierAddr + REG_IP_INT);

/* printk(KERN_NOTICE "====== irq handled entered %x.\n",INTSource);*/	

	if(!INTSource) 
	{
		spin_unlock(&ioc9010->port.lock);
		return IRQ_HANDLED;			/* Spurious interrupt, Kill it. */
	}

    INTFound = 0;
    for(i=0; i<6; i++)
    {
        if (INTSource & (1 << i*2)) 
        {
            INTFound = 1;
            break;
        }
    }

/*printk(KERN_NOTICE "====== irq handled entered.s=%x f=%d i=%d slot=%d\n",INTSource, INTFound, i, ioc9010->ipSlot);	*/

	/*
	 * Check whether the interrupt is for this IP card. If not, pass it on
	 * to anybody who is interested.
	 */
	if(INTFound)
	{
		if(ioc9010->ipSlot != i)
		{
			spin_unlock(&ioc9010->port.lock);
			return IRQ_NONE;			/* Not my interrupt, it is for other IP card. Quit with unhandled. */
		}
	}

	/*
	 * Check whether the interrupt is for this port. If not, pass it on
	 * to others.
	 */
	IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, INTERRUPT_REG_85156, &data);   //read interrupt back
	for (i=0; i<8; i++)
	{
		if(data & (1 << i))
		{
			port = i;
            break;
        }
    }

	if(port != ioc9010->portNo)
	{
		spin_unlock(&ioc9010->port.lock);
		return IRQ_NONE;				/* Not my interrupt, it is for other port. Quit with unhandled. */
	}

	/*
	 * Now, it is my interrupt. Handle it.
	 */
	DisableInterrupt(ioc9010, ioc9010->intLevel, 0);

	port = ioc9010->portNo;
	Hy85156_INTReason(ioc9010, &port, &reason);

/* printk(KERN_NOTICE "ttyHy%d irq handled reason=%x. \n",ioc9010->port.line, reason);*/
	
	/* Byte or break signal received */
	if ((reason & INTERRUPT_REASON_RXRDY) || ((reason & INTERRUPT_REASON_LSR) && (reason & INTERRUPT_REASON_LSR_RX_BREAK)))
		ioc9010_rx_chars(ioc9010);

	/* TX holding register empty - transmit a byte */
	if (reason & INTERRUPT_REASON_TXRDY)
		ioc9010_tx_chars(ioc9010);

	Hy85156_ResetInterrupt(ioc9010, ioc9010->portNo, reason);
	spin_unlock(&ioc9010->port.lock);
	return IRQ_HANDLED;
}

/*
 * interrupts disabled on entry
 */
static void ioc9010_stop_tx(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	int mask;

	mask = ioc9010->mask & (~ENABLE_THR_INT_85156);
	Hy85156_DisableInterrupt(ioc9010, ioc9010->portNo, mask);
}

/*
 * interrupts may not be disabled on entry
 */
static void ioc9010_start_tx(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	
	ioc9010->mask |= ENABLE_THR_INT_85156;
	Hy85156_EnableInterrupt(ioc9010, ioc9010->portNo, ioc9010->mask, 
							ioc9010->rcvTrigLevel, ioc9010->transTrigLevel);
}

/*
 * Interrupts enabled
 */
static void ioc9010_stop_rx(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	int mask;

	mask = ioc9010->mask & (~ENABLE_RHR_INT_85156);
	Hy85156_DisableInterrupt(ioc9010, ioc9010->portNo, mask);
}

/*
 * Set the modem control timer to fire immediately.
 */
static void ioc9010_enable_ms(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	mod_timer(&ioc9010->timer, jiffies);
}

static void ioc9010_rx_chars(struct ioc9010_uart_port *ioc9010)
{
	unsigned int ch, flag;
	long status;
	USHORT data;
	USHORT RXCounter;
	int i;

    do
    {
	   status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + RXCNT_85156*2, 1, &RXCounter);//read RFCNT register
	   if (RXCounter > 0)
	   {	
		  for(i=0; i<RXCounter; i++)
		  {
              status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + RHR_85156*2, 1, &data);//read data from buffer to value pointer
/*printk(KERN_NOTICE "=====p=%d rcv=%x ",ioc9010->port.line, data);*/	
		      if ((status == 0) && (data != 0))
              {
				 flag = TTY_NORMAL;

                 ch = (unsigned int) data & 0xFF;
		 		 ioc9010->port.icount.rx++;

	             status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//read receive status LSR
		 		 if (unlikely(data & (LSR_BREAK_85156 | LSR_FRAME_85156 | LSR_PARITY_85156 | LSR_OVERRUN_85156))) 
				 {
				 	if (data & LSR_BREAK_85156) 
					{
				 		ioc9010->port.icount.brk++;
				 		if (uart_handle_break(&ioc9010->port))
							continue;
				 	} else if (data & LSR_FRAME_85156)
				 		ioc9010->port.icount.frame++;
				 	else if (data & LSR_PARITY_85156)
				 		ioc9010->port.icount.parity++;
				 	if (data & LSR_OVERRUN_85156)
				 		ioc9010->port.icount.overrun++;

					data &= ioc9010->port.read_status_mask;
					if (data & LSR_BREAK_85156)
						flag = TTY_BREAK;
					else if (data & LSR_FRAME_85156)
						flag = TTY_FRAME;
					else if (data & LSR_PARITY_85156)
						flag = TTY_PARITY;
				 }

				 if (uart_handle_sysrq_char(&ioc9010->port, ch, NULL))
				 	continue;

				 if(ch != 0) uart_insert_char(&ioc9010->port, data, LSR_OVERRUN_85156, ch, flag);
 /*printk(KERN_NOTICE " char=%x flag=%d\n", ch, flag);*/	
             }
		  }
	   }
    }while((RXCounter > 0) && (status == 0));

/*	spin_lock(&ioc9010->port.lock);*/
	tty_flip_buffer_push(ioc9010->port.info->tty);
/*	tty_flip_buffer_push(ioc9010->port.state->port.tty);kernel2.6.37*/
/*	spin_unlock(&ioc9010->port.lock);*/
}

static void ioc9010_tx_chars(struct ioc9010_uart_port *ioc9010)
{
/*	struct circ_buf *xmit = &ioc9010->port.state->xmit;kernel2.6.37*/
	struct circ_buf *xmit = &ioc9010->port.info->xmit;
	unsigned char value[64];
	USHORT data;
	int status;

	if (ioc9010->port.x_char) 
	{
		value[0] = ioc9010->port.x_char;
		Hy85156_Write(ioc9010, ioc9010->portNo, value, 1);
		ioc9010->port.icount.tx++;
		ioc9010->port.x_char = 0;
		return;
	}

	/*
	 * Check the modem control lines before
	 * transmitting anything.
	 */
	ioc9010_mctrl_check(ioc9010);

	if (uart_circ_empty(xmit) || uart_tx_stopped(&ioc9010->port)) 
	{
		ioc9010_stop_tx(&ioc9010->port);
		return;
	}

	/*
	 * TX while bytes available
	 */
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + TXCNT_85156*2, 1, &data);//read TXCNT register	
	while (data < 40) 
	{
		value[0] = xmit->buf[xmit->tail];
		Hy85156_Write(ioc9010, ioc9010->portNo, value, 1);
/*printk(KERN_NOTICE "------p=%d sent=%x \n",ioc9010->port.line, value[0]);*/	
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		ioc9010->port.icount.tx++;
		if (uart_circ_empty(xmit))
			break;
		status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + TXCNT_85156*2, 1, &data);//read TXCNT register	
	}

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&ioc9010->port);

	if (uart_circ_empty(xmit))
		ioc9010_stop_tx(&ioc9010->port);
}


/*
 * Return TIOCSER_TEMT when transmitter is not busy.
 */
static unsigned int ioc9010_tx_empty(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	USHORT data;
	int status;

	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//read LSR register	
	return (data & THR_EMPTY_85156) ? TIOCSER_TEMT : 0;
}

static unsigned int ioc9010_get_mctrl(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	unsigned int mctrl = 0;
	USHORT data;
	int status;

	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + MSR_85156*2, 1, &data);//read MSR register

	if(status==0)
	{
		mctrl |= data & MSR_CTS_85156 ? TIOCM_CTS : 0;
		mctrl |= data & MSR_DSR_85156 ? TIOCM_DSR : 0;
		mctrl |= data & MSR_RI_85156 ? TIOCM_RI : 0;
		mctrl |= data & MSR_CD_85156 ? TIOCM_CAR : 0;
	}
	return mctrl;
}

static void ioc9010_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	USHORT data;

	IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//read MCR register

	if (mctrl & TIOCM_DTR)
		data |= MCR_DTR_85156;

	if (mctrl & TIOCM_RTS)
		data |= MCR_RTS_85156;

	if (mctrl & TIOCM_LOOP)
		data |= MCR_LOOP_85156;

	IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//write to MCR
}

/*
 * Interrupts always disabled.
 */
static void ioc9010_break_ctl(struct uart_port *port, int break_state)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	unsigned long flags;
	USHORT data;

	spin_lock_irqsave(&ioc9010->port.lock, flags);
	IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//read LCR register
	if (break_state == -1)
		data |= LCR_TXBREAK_85156;
	else
		data &= ~LCR_TXBREAK_85156;
	IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//write to LCR
	spin_unlock_irqrestore(&ioc9010->port.lock, flags);
}

static int ioc9010_startup(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	int retval = 0;

	spin_lock_irq(&ioc9010->port.lock);

	/* initialise com portNo to 9600,8,N,1 and no flow control */
	if(ioc9010->moduleType == SIGNATURE_8516)										/* for 8516 */
		Hy8516_ComsSetup(ioc9010, ioc9010->portNo, ioc9010->baud,
							ioc9010->parity, ioc9010->stopbits,
							ioc9010->databits, ioc9010->flowControl, 
							ioc9010->duplex, ioc9010->delay485);
	else																			/* for 8515 */
		Hy85156_ComsSetup(ioc9010, ioc9010->portNo, ioc9010->baud,
							ioc9010->parity, ioc9010->stopbits,
							ioc9010->databits, ioc9010->flowControl);

    /* Request the IRQ and link it to the Interrupt Handler */
	retval = request_irq(ioc9010->port.irq, ioc9010_isr, SA_INTERRUPT | SA_SHIRQ, "Hytec Serial INT Handler", ioc9010);
	if (retval)
	    {
	        printk(KERN_NOTICE "ttyHy%d - Int Handler can NOT assigned IRQ %d !\n", ioc9010->port.line, ioc9010->port.irq);
		}
	    else
	    {
	        printk(KERN_NOTICE "ttyHy%d - Interrupt Handler assigned to IRQ %d !\n", ioc9010->port.line, ioc9010->port.irq);
		}
	/*
	 * Enable interrupts
	 */
	retval = Hy85156_EnableInterrupt(ioc9010, ioc9010->portNo, ioc9010->mask, 
							ioc9010->rcvTrigLevel, ioc9010->transTrigLevel);
    
/*
	 * Enable modem status interrupts
	 */
	ioc9010_enable_ms(&ioc9010->port);
	spin_unlock_irq(&ioc9010->port.lock);

	return retval;
}

static void ioc9010_shutdown(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	int reason, rtport, status;
	USHORT data;
	/*
	 * Stop our timer.
	 */
	del_timer_sync(&ioc9010->timer);

	/*
	 * Reset the Tx and Rx FIFOS, disable the break condition
	 */
	Hy85156_Flush_Single(ioc9010, ioc9010->portNo, 1, 1);

	/*
	 * Clear all interrupts
	 */
	rtport = ioc9010->portNo;
	Hy85156_INTReason(ioc9010, &rtport, &reason);
	Hy85156_ResetInterrupt(ioc9010, ioc9010->portNo, reason);

	/*
	 * Disable interrupt on this port
	 */
   	data = 0; 
   	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) ioc9010->portNo * PORT_OFFSET + IER_85156*2, 1, &data);//write to FCTR
/*	Hy85156_DisableInterrupt(ioc9010, ioc9010->portNo, 0); here we cannot disable the whole thing. */

	/*
	 * Free the interrupt
	 */
	free_irq(ioc9010->port.irq, ioc9010);
    printk(KERN_NOTICE "ttyHy%d - irq handler (level %d) has been freed!\n", ioc9010->port.line, ioc9010->port.irq);
}

/*static void ioc9010_set_termios(struct uart_port *port, struct ktermios *termios,
		   						struct ktermios *old)* kernel 2.6.37*/
static void ioc9010_set_termios(struct uart_port *port, struct termios *termios,
		   						struct termios *old)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	unsigned long flags;

	/* Byte size.  */
	switch (termios->c_cflag & CSIZE) 
	{
	case CS5:
		ioc9010->databits = 5;
		break;
	case CS6:
		ioc9010->databits = 6;
		break;
	case CS7:
		ioc9010->databits = 7;
		break;
	case CS8:
	default:
		ioc9010->databits = 8;
		break;
	}

	/* Parity and stop bits.  */
	if (termios->c_cflag & CSTOPB)
		ioc9010->stopbits = 2;
	else
		ioc9010->stopbits = 1;

	if (termios->c_cflag & PARENB)
	{	
		if (termios->c_cflag & PARODD)
			ioc9010->parity = 'O';
		else
			ioc9010->parity = 'E';
	}
	else
		ioc9010->parity = 'N';

	/*
	 * Ask the core to calculate the baud rate.
	 */
	ioc9010->baud = uart_get_baud_rate(port, termios, old, 100, 921600);

	spin_lock_irqsave(&ioc9010->port.lock, flags);

	ioc9010->port.read_status_mask = LSR_OVERRUN_85156;
	if (termios->c_iflag & INPCK)
		ioc9010->port.read_status_mask |= LSR_FRAME_85156 | LSR_PARITY_85156;
	else
		ioc9010->port.read_status_mask &= ~(LSR_FRAME_85156 | LSR_PARITY_85156);
	
	if (termios->c_iflag & (BRKINT | PARMRK))
		ioc9010->port.read_status_mask |= LSR_BREAK_85156;
	else
		ioc9010->port.read_status_mask &= ~LSR_BREAK_85156;

	if (termios->c_iflag & IXON)								/* output software flow control */
		ioc9010->flowControl |= SOFTWARE_TRANS_FLOW_85156;
	else
		ioc9010->flowControl &= ~SOFTWARE_TRANS_FLOW_85156;

	if (termios->c_iflag & IXOFF)								/* input software flow control */
		ioc9010->flowControl |= SOFTWARE_RCV_FLOW_85156;
	else
		ioc9010->flowControl &= ~SOFTWARE_RCV_FLOW_85156;

	/*
	 * Characters to ignore
	 */
	ioc9010->port.ignore_status_mask = 0;
	if (termios->c_iflag & IGNPAR)
		ioc9010->port.ignore_status_mask |= LSR_FRAME_85156 | LSR_PARITY_85156;
	else
		ioc9010->port.ignore_status_mask &= ~(LSR_FRAME_85156 | LSR_PARITY_85156);

	if (termios->c_iflag & IGNBRK) 
	{
		ioc9010->port.ignore_status_mask |= LSR_BREAK_85156;
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns too (for real raw support).
		 */
		if (termios->c_iflag & IGNPAR)
			ioc9010->port.ignore_status_mask |= LSR_OVERRUN_85156;
	}else
	{
		ioc9010->port.ignore_status_mask &= ~(LSR_BREAK_85156 | LSR_OVERRUN_85156);
	}

	/*
	 * Receive interrupt 
	 */
	if (termios->c_cflag & CREAD)
		ioc9010->mask |= ENABLE_RHR_INT_85156;
	else
		ioc9010->mask &= ~ENABLE_RHR_INT_85156;


	del_timer_sync(&ioc9010->timer);

	/*
	 * Update the per-port timeout.
	 */
	uart_update_timeout(&ioc9010->port, termios->c_cflag, ioc9010->baud);

	/*
	 * disable interrupts and drain transmitter
	 */
	Hy85156_DisableInterrupt(ioc9010, ioc9010->portNo, 0);

	/*
	 * Reset the Tx and Rx FIFOS, disable the break condition
	 */
	Hy85156_Flush_Single(ioc9010, ioc9010->portNo, 1, 1);

	/* set up the port */
	if(ioc9010->moduleType == SIGNATURE_8516)										/* for 8516 */
		Hy8516_ComsSetup(ioc9010, ioc9010->portNo, ioc9010->baud,
							ioc9010->parity, ioc9010->stopbits,
							ioc9010->databits, ioc9010->flowControl, 
							ioc9010->duplex, ioc9010->delay485);
	else																			/* for 8515 */
		Hy85156_ComsSetup(ioc9010, ioc9010->portNo, ioc9010->baud,
							ioc9010->parity, ioc9010->stopbits,
							ioc9010->databits, ioc9010->flowControl);

	Hy85156_EnableInterrupt(ioc9010, ioc9010->portNo, ioc9010->mask, 
							ioc9010->rcvTrigLevel, ioc9010->transTrigLevel);

	if (UART_ENABLE_MS(&ioc9010->port, termios->c_cflag))
		ioc9010_enable_ms(&ioc9010->port);

	spin_unlock_irqrestore(&ioc9010->port.lock, flags);
}

static const char *ioc9010_type(struct uart_port *port)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	return ioc9010->port.type == PORT_IOC9010 ? "IOC9010" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'.
 */
static void ioc9010_release_port(struct uart_port *port)
{
	/* we don't need to release memory region here. It's done by driver ioc9010_serial_exit. */
	return;
}

/*
 * Request the memory region(s) being used by 'port'.
 */
static int ioc9010_request_port(struct uart_port *port)
{
	/* we don't need to request memory region here. It's already done during the PCI device probe. */
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	if(ioc9010->port.mapbase != 0) return 0;
	return -EBUSY;
}

/*
 * Configure/autoconfigure the port.
 */
static void ioc9010_config_port(struct uart_port *port, int flags)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;

	if (flags & UART_CONFIG_TYPE && (ioc9010_request_port(&ioc9010->port) == 0))
		ioc9010->port.type = PORT_IOC9010;
}

/*
 * we don't want the core code to modify any port params.
 */
static int ioc9010_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	/* we don't want the core code to modify any port params */
	return -EINVAL;
}



/* UART operations
 *
 * Details on these functions can be found in Documentation/serial/driver
 */
static struct uart_ops ioc9010_uart_pops = {
	.tx_empty       = ioc9010_tx_empty,
	.set_mctrl      = ioc9010_set_mctrl,
	.get_mctrl      = ioc9010_get_mctrl,
	.stop_tx		= ioc9010_stop_tx,
	.start_tx       = ioc9010_start_tx,
	.stop_rx		= ioc9010_stop_rx,
	.enable_ms      = ioc9010_enable_ms,
	.break_ctl      = ioc9010_break_ctl,
	.startup		= ioc9010_startup,
	.shutdown       = ioc9010_shutdown,
	.set_termios    = ioc9010_set_termios,
	.type   		= ioc9010_type,
	.release_port   = ioc9010_release_port,
	.request_port   = ioc9010_request_port,
	.config_port    = ioc9010_config_port,
	.verify_port    = ioc9010_verify_port,
};

static struct uart_driver ioc9010_uart_driver;

/*
 * Setup the IOC9010 serial ports.
 *
 * Note also that we support "console=ttyHyxyz" where "x" is 00~31, "y" is 0~6, "z" is 0~8.
 */
static void __init ioc9010_init_ports(void)
{
	static int first = 1;
	int i, j, k;
	struct IOC9010_dev *ioc9010dev; 
	int serialCardNos = 0, cardNo = 0;
	USHORT data;
	long status;
 
	if (!first)										/* only need once */
		return;
	first = 0;

	/* Begin */
	printk(KERN_NOTICE "HYTEC Linux UART Device - initialising port ... \n");	

	serialCardNos = buildHytecPciDeviceList();

	if(serialCardNos > 0)
	{

		IOC9010_UART_MINOR = serialCardNos * 8;		/* every 8515/8516 has 8 serial ports */

		/* ????? Somehow, tty device driver doesn't like dynamic major number ???? */
		/* Dynamically Allocate one instead...*/
		/*retval = alloc_chrdev_region(&dtMajorMinor, 0, MAXDEVNUM, "IOC9010");
		IOC9010_UART_MAJOR = MAJOR(dtMajorMinor);*/
	
		ioc9010dev = pciDevList;
		k = 0;
		cardNo = 0;

		while (ioc9010dev != NULL)
		{
			for (i=0; i<ioc9010dev->ipSlotNum; i++)
			{
				if((ioc9010dev->ipCard[i] == 0x8515) || (ioc9010dev->ipCard[i] == 0x8516))
				{
					for (j=0; j<8; j++)
					{
						/* ioc9010 port extension init */
						ioc9010_ports[k].pcidev = ioc9010dev;
						ioc9010_ports[k].device_ID = ioc9010dev->device_ID;
						ioc9010_ports[k].virtureMemBaseAddr = ioc9010dev->virtureMemBaseAddr;
						ioc9010_ports[k].carrierAddr = ioc9010dev->carrierAddr;
						ioc9010_ports[k].plxLocalConfigBAR = ioc9010dev->plxLocalConfigBAR;
						ioc9010_ports[k].ipSlot = i;
						ioc9010_ports[k].portNo = j;
						ioc9010_ports[k].moduleType = ioc9010dev->ipCard[i];

						init_timer(&ioc9010_ports[k].timer);
						ioc9010_ports[k].timer.function = ioc9010_timeout;
						ioc9010_ports[k].timer.data = (unsigned long)&ioc9010_ports[i];

						/* initialise uart default */
						ioc9010_ports[k].baud = 9600;
						ioc9010_ports[k].parity = 'N';
						ioc9010_ports[k].stopbits = 1;
						ioc9010_ports[k].databits = 8;
						ioc9010_ports[k].flowControl = 0;
						if(ioc9010_ports[k].moduleType == 0x8516) 
						{
							ioc9010_ports[k].duplex = 2;				/* initial auto half duplex mode without echo */
							ioc9010_ports[k].withEcho = 0;				/* initial setting as without echo */
							ioc9010_ports[k].delay485 = 5;				/* set to delay 5 bits time for half duplex mode */
						}
						ioc9010_ports[k].mask = 0;						/* default to no interrupt */
						ioc9010_ports[k].rcvTrigLevel = 1;				/* trigger interrupt at receiving 1 chars */
						ioc9010_ports[k].transTrigLevel = 5;			/* trigger interrupt when transmit FIFO reaches this level */
						ioc9010_ports[k].vector = 0xC0 + cardNo;		/* not very nice to hard code the vector number */
						ioc9010_ports[k].flag = 0;
						
						/* uart port init */
						ioc9010_ports[k].port.ops = &ioc9010_uart_pops;
						ioc9010_ports[k].port.iobase = (unsigned long)(ioc9010dev->iocMemoryAddr + IP_IO_BASE_ADDR + IP_A_IO_BASE_ADDR + (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * ioc9010_ports[k].ipSlot);
						ioc9010_ports[k].port.membase = (unsigned char*)(ioc9010dev->iocMemoryAddr + (IP_B_MEMORY_BASE_ADDR - IP_A_MEMORY_BASE_ADDR) * ioc9010_ports[k].ipSlot);
						ioc9010_ports[k].port.irq = ioc9010dev->carrierIRQ;
						ioc9010_ports[k].port.fifosize = FIFO_SIZE;
						/*ioc9010dev->carrierNum * 100 + i * 10 + j;*/
						ioc9010_ports[k].port.type = PORT_IOC9010;
						ioc9010_ports[k].port.iotype = SERIAL_IO_MEM;
						ioc9010_ports[k].port.line = k;

						uart_add_one_port(&ioc9010_uart_driver, &ioc9010_ports[k].port);
						k++;
					}

					/* 
					 * The following sets up the flag to notify the PCI device driver if loaded
 					 * not to response any interrupts raised by the serial cards. This driver is
					 * taking care of them.
					 */
					status = IOC9010IPRead(ioc9010_ports[k-1].virtureMemBaseAddr, ioc9010_ports[k-1].ipSlot, TIMER_COUNTER_LS_85156, &data);					//read timer controller reg
					data = (data & 0x0F) | 0xAA;      
					status = IOC9010IPWrite(ioc9010_ports[k-1].virtureMemBaseAddr, ioc9010_ports[k-1].ipSlot, TIMER_COUNTER_LS_85156, data);					//set 0xA0 flag to notify PCI driver is loaded that I am here in control

					/* 
					 * Note, interrupt vector is hard coded starting from 
					 * 0xC0 plus card sequence number. The configuration
					 * is only needed per IP. It does it at the last one.
					 */
					Hy85156_Configure(&ioc9010_ports[k-1], 0xC0+cardNo, 0);											/* initialise 8515/8516 with interrupt, do not use delay */
					cardNo++;					
				}
			}
			ioc9010dev = ioc9010dev->next;
		}

		printk(KERN_NOTICE "HYTEC Linux UART Device - %d serial devices have been registered. \n", IOC9010_UART_MINOR);
	}else
		printk(KERN_NOTICE "HYTEC Linux UART Device - no serial device found. \n");
}

#ifdef CONFIG_SERIAL_IOC9010_CONSOLE

static void ioc9010_console_putchar(struct uart_port *port, int ch)
{
	struct ioc9010_uart_port *ioc9010 = (struct ioc9010_uart_port *)port;
	int status;
	unsigned char buffer[2];

	buffer[0] = ch & 0xFF;
	status = Hy85156_Write(ioc9010, ioc9010->portNo, buffer, 1);
}

/*
 * Interrupts are disabled on entering
 */
static void ioc9010_console_write(struct console *co, const char *s, unsigned int count)
{
	struct ioc9010_uart_port *ioc9010 = &ioc9010_ports[co->index];
	unsigned int old_ien, status;

	/*
	 *	First, save IEN and then disable interrupts
	 */

	uart_console_write(&ioc9010->port, s, count, ioc9010_console_putchar);

	/*
	 *	Finally, wait for transmitter to become empty
	 *	and restore IEN
	 */
}

static int __init ioc9010_console_setup(struct console *co, char *options)
{
	struct ioc9010_uart_port *ioc9010;
	int baud = 9600;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	if (co->index == -1 || co->index >= IOC9010_UART_MINOR)
		co->index = 0;
	ioc9010 = &ioc9010_ports[co->index];

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);

	return uart_set_options(&ioc9010->port, co, baud, parity, bits, flow);
}

static struct console ioc9010_console = {
	.name		= "ttyHy",
	.write		= ioc9010_console_write,
	.device		= uart_console_device,
	.setup		= ioc9010_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &ioc9010_uart_driver,
};

static int __init ioc9010_rs_console_init(void)
{
	ioc9010_init_ports();
	register_console(&ioc9010_console);
	return 0;
}

console_initcall(ioc9010_rs_console_init);

#define IOC9010_CONSOLE	&ioc9010_console
#else
#define IOC9010_CONSOLE	NULL
#endif


static struct uart_driver ioc9010_uart_driver = {
	.owner  		= THIS_MODULE,
	.driver_name    = "ioc9010_uart",
	.dev_name       = "ttyHy",
	.major  		= IOC9010_UART_MAJOR,
	.minor  		= MAX_UART_NUM,
	.nr     		= MAX_UART_NUM,
	.cons			= IOC9010_CONSOLE
};

/* Note, IOC9010 serious is not platform device so we don't need platform related stuff */

static int __init ioc9010_serial_init(void)
{
	int status;

	printk(KERN_INFO "IOC9010: Entered serial init.\n");

	status = uart_register_driver(&ioc9010_uart_driver);
	if (status == 0)
		ioc9010_init_ports();
	else
		uart_unregister_driver(&ioc9010_uart_driver);
	return status;
}

static void __exit ioc9010_serial_exit(void)
{
	int i, status;
	USHORT data;

	/* Destroy the driver */
	for (i = 0; i < IOC9010_UART_MINOR; ++i) 
	{
		if (&ioc9010_ports[i]) 
		{
			/* 
			 * The following clears up the flag to notify the PCI device driver if loaded
			 * that this driver finishes its mission. Note, for each IP, we only need to 
			 * do this once.
			 */
			if((1%8) == 0) 
			{
				status = IOC9010IPRead(ioc9010_ports[i].virtureMemBaseAddr, ioc9010_ports[i].ipSlot, TIMER_COUNTER_LS_85156, &data);					//read timer controller reg
				data = data & 0x0F;      
				status = IOC9010IPWrite(ioc9010_ports[i].virtureMemBaseAddr, ioc9010_ports[i].ipSlot, TIMER_COUNTER_LS_85156, data);					//clear 0xA0 flag to notify PCI driver that I finished using the serial
			}
			uart_remove_one_port(&ioc9010_uart_driver, &ioc9010_ports[i].port);
		}
	}

	uart_unregister_driver(&ioc9010_uart_driver);

	/* remove registered PCI devices */
	removePciDevice();
}

module_init(ioc9010_serial_init);
module_exit(ioc9010_serial_exit);



/******************************************************************************
 *																			  *
 *                      INTERNAL FUNCTION AREA								  *
 *     																		  *
 *****************************************************************************/


/******************************************************************************
 *
 * Function   :  buildHytecPciDeviceList
 * Description:  Identify Hytec 8515 and 8516 devices and build device list
 * Argument:	 none
 * Return:		 >0: number of serial IP cards found. 0: no 8515/8516 found.
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
int buildHytecPciDeviceList(void)
{
    int             status;
    int             devCount;
    int             serialCount;
    struct pci_dev *pPciDev;


	devCount = 0;
	serialCount = 0;

	//get the first PCI device
	pPciDev = pci_get_device(
        PCI_ANY_ID,
        PCI_ANY_ID,
        NULL
        );

	while(pPciDev)
	{
		if(checkHytecPciDevice(pPciDev))
		{
            // Add the device to the device list
            status = addPciDevice(pPciDev);
            if (status != 0)
            {
                // Increment number of devices found
                devCount++;
				serialCount += status;
            }
		}

		// Jump to next device
        pPciDev = pci_get_device(
            PCI_ANY_ID,
            PCI_ANY_ID,
            pPciDev
            );
	}

	printk(KERN_NOTICE "HYTEC Linux UART Device: found %d PCI device(s) and %d serial IP(s).\n", devCount, serialCount);
    return serialCount;
}

/******************************************************************************
 *
 * Function   :  checkHytecPciDevice
 * Description:  Check a new device whether is a Hytec device
 * Argument:	 pPciDev - pointer of pci_dev structure
 * Return:		 1: yes, Hytec device. 0: not Hytec device
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
int checkHytecPciDevice(struct pci_dev *pPciDev)
{
	int found = 0;
	
	struct pci_device_id* pciid = ids;

	while(pciid->vendor != 0)
	{
		if((pPciDev->vendor == pciid->vendor) && (pPciDev->device == pciid->device))
		{
			return 1;
			break;
		}
		pciid++;
	}

	return found;
}



/******************************************************************************
 *
 * Function   :  addPciDevice
 * Description:  Add Hytec PCI device to device list
 * Argument:	 pPciDev - pointer of pci_dev structure
 * Return:		 0: nothing added. >0: number of serial IP(s) ound
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
int addPciDevice(struct pci_dev *pPciDev)
{
	u32 baseAddressVariable;
	u32 carrierBoardBaseAddress;
	int serialNum = 1;									/* default serial number of the carrier is 1 */

	u16 endIDBaseAddr;
	u32 ipOffset;
	u16 dataRead;
	u16 cardCount = 0;
	u16 serialFound = 0;

	struct IOC9010_dev* ioc9010dev; 
	struct IOC9010_dev* devtemp; 

	/* Create a device and added to the device chain */
	ioc9010dev = kmalloc(sizeof(struct IOC9010_dev), GFP_KERNEL);
	memset(ioc9010dev, 0, sizeof(struct IOC9010_dev));	/* zero everything */

	ioc9010dev->next = pciDevList;
	devtemp = pciDevList;
	pciDevList = ioc9010dev;

	ioc9010dev->vendor_ID = pPciDev->vendor;
	ioc9010dev->device_ID = pPciDev->device;
	ioc9010dev->pci_bus = pPciDev->bus->number;
	ioc9010dev->pci_dev = PCI_SLOT(pPciDev->devfn);
	ioc9010dev->pci_func = PCI_FUNC(pPciDev->devfn);
	ioc9010dev->carrierIRQ = pPciDev->irq;

	switch (ioc9010dev->device_ID)
    {
    case 0x9010:
    case 0x6335:
    case 0x7002:

    	/* Map the Base Address for the 9010's PCI Carrier Card Registers */
    	/* -------------------------------------------------------------- */
    	pci_read_config_dword(pPciDev, PCI_BASE_ADDRESS_2, &carrierBoardBaseAddress);
	
    	/* These are I/O, hence LSB is only set to indicate an I/O Region */
    	carrierBoardBaseAddress--;			//bit0 = 1 means it is I/O base, so mask bit0
	
    	/* Update Log about it all went */
    	printk(KERN_NOTICE "%x Test - Base Add2 - 0x%8X \n",ioc9010dev->device_ID, carrierBoardBaseAddress);
    	if(!request_region(carrierBoardBaseAddress,64,"IOC9010 Carrier"))
    	{ 
    		printk(KERN_NOTICE "%x Test - Carrier region is also used by another driver!\n", ioc9010dev->device_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - Carrier Mapped !\n", ioc9010dev->device_ID);
			ioc9010dev->bar2State = 1;														//BAR2 mapped successfully
    	}

    	/* store carrier register base address */
    	ioc9010dev->carrierAddr = carrierBoardBaseAddress;

		/* check duplication card number settings for 7002 or 6335 */
		if((ioc9010dev->device_ID == 0x7002) || (ioc9010dev->device_ID == 0x6335))	/* get the serial number of the carrier */
		{
			serialNum = inw(ioc9010dev->carrierAddr + 0x04) + 1;
			while (devtemp != NULL)
			{
				if ((devtemp->carrierNum) == serialNum)
				{ 
    				printk(KERN_NOTICE "%x Test - Duplicated carrier ID %d found (Not registered). Please check carrier ID jumpers.\n", devtemp->device_ID, serialNum);
					pciDevList = ioc9010dev->next;
					kfree(ioc9010dev);							/* free the current */
					return 0;
				}
				devtemp = devtemp->next;
			}
    		printk(KERN_NOTICE "%x Test - Carrier ID=%d.\n", ioc9010dev->device_ID, serialNum);
		}

		devtemp = 0;
		ioc9010dev->carrierNum = serialNum;

    	/* Map the Base Address for the IP Cards */
    	pci_read_config_dword(pPciDev, PCI_BASE_ADDRESS_3, &baseAddressVariable);
    	printk(KERN_NOTICE "%x Test - Base Add3 - 0x%8X \n", ioc9010dev->device_ID, baseAddressVariable);
    	if(!request_mem_region(baseAddressVariable,0xE00E00,"IOC9010 IPs"))
    	{
    		printk(KERN_NOTICE "%x Test - Memory is also used by another driver!\n", ioc9010dev->device_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - Memory for IPs Mapped !\n", ioc9010dev->device_ID);
			ioc9010dev->bar3State = 1;														//BAR3 mapped successfully
    	}

		ioc9010dev->iocMemoryAddr = baseAddressVariable;			//save it for releasing
    	ioc9010dev->virtureMemBaseAddr = ioremap(baseAddressVariable,0xE00E00);	

    	/* Map the Base Address for the PLX Local Configuration Registers I/O region*/
    	pci_read_config_dword(pPciDev, PCI_BASE_ADDRESS_1, &carrierBoardBaseAddress);
	
    	/* These are I/O, hence LSB is only se 0x    E000 t to indicate an I/O Region */
    	carrierBoardBaseAddress--;				//bit0 = 1 means I/O space, should mask this bit
    	printk(KERN_NOTICE "%x Test - Base Add1 - 0x%8X \n",ioc9010dev->device_ID, carrierBoardBaseAddress);
		if(!request_region(carrierBoardBaseAddress,80,"Hytec PLX Local"))
    	{ 
    		printk(KERN_NOTICE "%x Test - PLX Local region is also used by another driver!\n",ioc9010dev->device_ID);
    	}
    	else
    	{
    		printk(KERN_NOTICE "%x Test - PLX Local Config Registers Mapped !\n",ioc9010dev->device_ID);
			ioc9010dev->bar1State = 1;														//BAR1 mapped successfully
    	}
		
		ioc9010dev->plxLocalConfigBAR = carrierBoardBaseAddress;

		/* Now, figure out is there a serial device or not */
		endIDBaseAddr = (ioc9010dev->device_ID == 0x9010) ? IP_F_ID_BASE_ADDR : ((ioc9010dev->device_ID == 0x6335) ? IP_B_ID_BASE_ADDR : IP_A_ID_BASE_ADDR);
		/* ipslot number: IOC9010:6, 6335:2, 7220:1 */
		ioc9010dev->ipSlotNum = (ioc9010dev->device_ID == 0x9010) ? 6 : ((ioc9010dev->device_ID == 0x6335) ? 2 : 1);

		/* Scan for IP Slots - Step Through the Possible ID Address Spaces */
    	cardCount = 0;
		serialFound = 0;
		for (ipOffset  = IP_A_ID_BASE_ADDR; ipOffset <= endIDBaseAddr; ipOffset += (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR)	)
    	{
    		dataRead = ioread16(ioc9010dev->virtureMemBaseAddr + IP_IO_BASE_ADDR + ipOffset + ID_MODEL_NUMBER); 
    		
    		printk(KERN_NOTICE "9010 Test - Slot %c ID Found - %4X !\n",'A'+cardCount, dataRead);

    		if (dataRead != 0) ioc9010dev->ipCard[cardCount] = dataRead;
			if ((dataRead == 0x8515) || (dataRead == 0x8516)) serialFound++;
    		cardCount++;
    	}
		
    	
		/* Check the PCI Carrier Interrupts */
    	/* -------------------------------- */
        pci_read_config_byte(pPciDev, PCI_INTERRUPT_PIN, &ioc9010dev->carrierIRQ);
	
    	// Update Log about it all went
    	printk(KERN_NOTICE "%x Test - IRQ Pin - %d \n", ioc9010dev->device_ID, ioc9010dev->carrierIRQ);
    	ioc9010dev->carrierIRQ = pPciDev->irq;
		ioc9010dev->ppciDev = pPciDev;						/* store pci device reference */
    	printk(KERN_NOTICE "%x Test - IRQ level- %d\n", ioc9010dev->device_ID, ioc9010dev->carrierIRQ);

		break;

    default:		/* NOTE: this is not fully support by the time (11/01/2011), need more work and thought!! */
        break;
	}

    // Enable the device
    if (pci_enable_device(pPciDev) != 0)
    {
        printk(KERN_NOTICE "%x Test - ERROR - PCI device enable failed !\n",ioc9010dev->device_ID);
		serialFound = 0;	/* if pci device cannot be enabled, no matter what, quit! */
    }

	if(!serialFound)		/* if no serial device found, no need to have this pci device registered */
	{
	    switch (ioc9010dev->device_ID)
	    {
	    case 0x9010:
	    case 0x6335:
	    case 0x7002:
    	
	        /* Tidy Up for Stuff Associated with Carrier Board */
			/* ----------------------------------------------- */
			/* Un Map Memory */ 
			if(ioc9010dev->bar3State == 1)
			{
				iounmap(ioc9010dev->virtureMemBaseAddr);	
				release_mem_region(ioc9010dev->iocMemoryAddr,0xE00E00);
			}
			if(ioc9010dev->bar1State == 1) release_region(ioc9010dev->plxLocalConfigBAR,80);
			if(ioc9010dev->bar2State == 1) release_region(ioc9010dev->carrierAddr,64);

	        break;

	    default:
	        break;

	    }

		printk(KERN_NOTICE "HYTEC Linux UART Device - The device is removed due to no serial found.\n");

		pciDevList = ioc9010dev->next;					/* retrieve next */
		kfree(ioc9010dev);							/* free the current */
	}
	
    return serialFound;
}




/******************************************************************************
 *
 * Function   :  removePciDevice
 * Description:  Remove Hytec PCI device from device list
 * Argument:	 none
 * Return:		 0: success
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
int removePciDevice(void)
{
	struct IOC9010_dev* ioc9010dev = pciDevList; 
	struct IOC9010_dev* devtemp; 

	/* clean mapping stuffs and pci devices */
	while (ioc9010dev != NULL)
	{
	    switch (ioc9010dev->device_ID)
	    {
	    case 0x9010:
	    case 0x6335:
	    case 0x7002:
    	
	        /* Tidy Up for Stuff Associated with Carrier Board */
			/* ----------------------------------------------- */
			/* Un Map Memory */ 
			if(ioc9010dev->bar3State == 1)
			{
				iounmap(ioc9010dev->virtureMemBaseAddr);	
				release_mem_region(ioc9010dev->iocMemoryAddr,0xE00E00);
			}
			if(ioc9010dev->bar1State == 1) release_region(ioc9010dev->plxLocalConfigBAR,80);
			if(ioc9010dev->bar2State == 1) release_region(ioc9010dev->carrierAddr,64);

	        break;

	    default:
	        break;
	    }

		printk(KERN_NOTICE "HYTEC Linux UART Device - The remove is called \n");

		pci_dev_put(ioc9010dev->ppciDev);			/* decrease the counter for the kernel? */
		devtemp = ioc9010dev->next;					/* retrieve next */
		kfree(ioc9010dev);							/* free the current */
		ioc9010dev = devtemp;
	}
	return 0;
}

/******************************************************************************
 *
 * Function:	IOC9010IPRead														
 * Description:	Executes a single action of 16 bit read from IP card registers.
 * Arguements:	memoryBaseAddress = memory base address      	
 *          	ipslot = IP card slot number 0 ~ 5												
 *				add  = register address															
 *				*data = pointer to hold the register value										
 * Return:		< 0: error code 																
 *				= 0: if successful
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
long IOC9010IPRead(void *memoryBaseAddress, int ipslot, USHORT add, USHORT *data)
{
    *data = ioread16(memoryBaseAddress + IP_IO_BASE_ADDR + IP_A_IO_BASE_ADDR + (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * ipslot + add);
	return 0;
}

/******************************************************************************
 *
 * Function:	IOC9010IPWrite														
 * Description:	Executes a single action of 16 bit write to IP card registers.
 * Arguements:	memoryBaseAddress = memory base address      	
 *          	ipslot = IP card slot number 0 ~ 5												
 *				add  = register address															
 *				*data = pointer to hold the register value										
 * Return:		< 0: error code 																
 *				= 0: if successful
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
long IOC9010IPWrite(void *memoryBaseAddress, int ipslot, USHORT add, USHORT data)
{
    iowrite16(data, memoryBaseAddress + IP_IO_BASE_ADDR + IP_A_IO_BASE_ADDR + (IP_B_ID_BASE_ADDR - IP_A_ID_BASE_ADDR) * ipslot + add);
	return 0;
}

/******************************************************************************
 *
 * Function:	IOC9010IPMemoryRead														
 * Description:	Executes a single action of 16 bit read from IP card memory.
 * Arguements:	memoryBaseAddress = memory base address      	
 *          	ipslot = IP card slot number 0 ~ 5												
 *				add  = register address															
 *				*data = pointer to hold the register value										
 * Return:		< 0: error code 																
 *				= 0: if successful
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
long IOC9010IPMemoryRead(void *memoryBaseAddress, int ipslot, USHORT add, int len, USHORT *data)
{
	int i;
    for (i=0; i<len; i++)
    {	
    	*(data + i) = ioread16(memoryBaseAddress + (IP_B_MEMORY_BASE_ADDR - IP_A_MEMORY_BASE_ADDR) * ipslot + add + i);
	}
	return 0;
}

/******************************************************************************
 *
 * Function:	IOC9010IPMemoryWrite														
 * Description:	Executes a single action of 16 bit write to IP card memory.
 * Arguements:	memoryBaseAddress = memory base address      	
 *          	ipslot = IP card slot number 0 ~ 5												
 *				add  = register address															
 *				*data = pointer to hold the register value										
 * Return:		< 0: error code 																
 *				= 0: if successful
 *-----------------------------------------------------------------------------
 * Author:		 Jim S Chen
 * Start:		 11/01/2011
 * Version:		 1.0
 *-----------------------------------------------------------------------------
 * Revision:
 *
 ******************************************************************************/
long IOC9010IPMemoryWrite(void *memoryBaseAddress, int ipslot, USHORT add, int len, USHORT *data)
{
	int i;
    for (i=0; i<len; i++)
    {	
	    iowrite16(*(data + i), memoryBaseAddress + (IP_B_MEMORY_BASE_ADDR - IP_A_MEMORY_BASE_ADDR) * ipslot + add + i);
	}
	return 0;
}

/*----------------------------------------------------------------------------------*/
/*Function:		Configure 85156 IP card												*/
/*----------------------------------------------------------------------------------*/
/*Description:	85156 IP configuration routine according to the settings.			*/
/*				Note: if use multiple channels, this is used for the first channel	*/
/*				only																*/	
/*Prototype:	long Hy85156_Configure(int vmeslot, char ipslot, char vector,		*/
/*				USHORT delay)														*/
/*Arguements:	vmeslot = The number of the VME carrier card on which the IP card	*/
/*					resides. The carrier must be initialised before this routine.	*/
/*					This param is only for VME.										*/
/*																					*/
/*				ipslot = IP card slot number 0 ~ 3 for VME, 0 ~ 5 for IOC			*/
/*																					*/
/*				vector = The interrupt vector number [1..255] designated to this IP	*/
/*					card. The card will not work without interrupts. Different		*/
/*					physical 8515/8516 cards may share the same interrupt number.	*/
/*					segment of the interrupt vector. only the upper 8 bits			*/
/*																					*/
/*				delay = For Positive Values -										*/
/*					Determine the number of clock ticks between timer interrupts.	*/
/*					The clock frequency is 125 KHz. The driver checks all active	*/
/*					channels of the serial card for characters that need to be read */	
/*					or written at every interrupt delay.							*/
/*																					*/
/*					For Negative Values -											*/
/*					The modulus of the Value sets the depth at which the RX FIFO	*/
/*					is filled and/or the TX FIFO is emptied before the serail IC	*/
/*					generates an Interrupt. Once an interrupt is detected, the		*/
/*					driver checks all active channels of the serial card for		*/
/*					characters that need to be read or written at every interrupt	*/
/*					delay.															*/
/*																					*/
/*					The modulus of intdelay should not be chosen to be too big for	*/
/*					a given data rate, keeping in mind that all channels have 64	*/
/*					byte FIFOs.														*/
/*																					*/
/*																					*/
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
long Hy85156_Configure(struct ioc9010_uart_port *ioc9010, char vector, int delay)
{
	USHORT data;
	long status;
    int i;
    		
	status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, CSR_85156, &data);					//read CSR back
	data = data & 0x0100;      
	status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, CSR_85156, data);					//reset all UARTs

    for(i=0; i<8; i++)
    {
    	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + EFR_85156*2, 1, &data);		//read MCR 
        data = data | 0x10;				//enable enhanced functions
    	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + EFR_85156*2, 1, &data);	//write to MCR
    	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + MCR_85156*2, 1, &data);		//read MCR 
        data = data & (~0x40);			//disable intrared
    	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + MCR_85156*2, 1, &data);	//write to MCR
    }

	if (vector != 0)																	//set up interrupt path
	{
		//enable IP interrupt
		status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, CSR_85156, &data);	            //read IP CSR back
		data = data | ENABLE_INTERRUPT_85156;      
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, CSR_85156, data);	            //enable IP interrupt		
	}

	//set up vector
	data = (USHORT) (vector << 8);
	status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, INTERRUPT_REG_85156, data);		//configure interrupt vector

/* 
 * We are not going use the timer on 8515/8516 IP, so don't bother. Bothering it
 * would screw up the locking mechanism between the PCI and UART drivers
 */
/*	if (delay > 0)
	{
		data = (USHORT) (delay & 0x00FF);
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_COUNTER_LS_85156, data);	//set timer counter LS
		data = (USHORT) ((delay & 0xFF00) >> 8);
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_COUNTER_MS_85156, data);	//set timer counter MS
		data = 0x0003;																	//internal clock, re-trigger, start timer, enable timer
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_CONTROL_85156, data);	//set timer control register
	}else
	{
		data = 0;
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_COUNTER_LS_85156, data);	//clear timer counter
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_COUNTER_MS_85156, data);	//clear timer counter
		data = 0;																		//internal clock, re-trigger, start timer, enable timer
		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, TIMER_CONTROL_85156, data);	//clear timer control register
	}
*/
	return status;																		//return status
}


/*----------------------------------------------------------------------------------*/
/*Function:		Configure 85156 COM Port											*/
/*----------------------------------------------------------------------------------*/
/*Description:	set up one 85156 COM portNo											*/
/*Prototype:	Hy85156_ComsSetup(int portNo, USHORT baud, char parity,				*/
/*				int numstopbits, int numdatabits, char flowctr)						*/
/*Arguements:	portNo = serial channel 0 ~ 7											*/
/*				baud = baud rate divisor. This can be the follwoing number only		*/
/*					baud setting	value to register	internal loopback			*/
/*					------------------------------------------------------			*/
/*						100			2304 (0x0900)			1						*/
/*						400			2304 (0x0900)			0						*/
/*						600			 384 (0x0180)			1						*/
/*						1200		 192 (0x00C0)			1						*/
/*						2400		 128 (0x0080)			0						*/
/*						4800		 192 (0x00C0)			0						*/
/*						9600		  96 (0x0060)			0						*/
/*						19200		  48 (0x0030)			0						*/
/*						38400		  24 (0x0018)			0						*/
/*						57600		   4 (0x0004)			1						*/
/*						76800		  12 (0x000C)			0						*/
/*						115200		   2 (0x0002)			1						*/
/*						153600		   6 (0x0006)			0						*/
/*						230400		   4 (0x0004)			0						*/
/*						460800		   2 (0x0002)			0						*/
/*						921600		   1 (0x0001)			0						*/
/*				parity = parity setting												*/
/*					"N"	no parity													*/
/*					"O"	odd parity													*/
/*					"E"	even parity													*/
/*				numdatabits = data bits												*/
/*					=5, 6, 7, 8	for 5,6,7,8 bits									*/
/*				numstopbits = sstop bits											*/
/*					1	1 stop bit													*/
/*					4	1 1/2 stop bit if data bit = 5								*/
/*					4	2 stop bit if data bit = 6,7,8								*/
/*				flowctr = flow control byte											*/
/*					0 0 X X X X X X													*/
/*					[_]	| | [_____]													*/
/*					 |	| |	   |													*/
/*			   no used	| |	software flow control									*/
/*						| | 0 0 0 0	no software flow control						*/
/*						| | 1 1 0 0	transmit Xon1 and Xon2 / Xoff1 and Xoff2		*/
/*						| |	0 0	1 1	receiver Xon1 and Xon2 / Xoff1 and Xoff2		*/
/*						| |	1 1 1 1 both transmit and receiver Xon1,2 / Xoff1,2		*/
/*						| =0 disable hardware flow control, =1 enable				*/
/*					    =0 use RTS/CTS, =1 use DTR/DSR pair							*/
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
long Hy85156_ComsSetup(struct ioc9010_uart_port *ioc9010, int portNo, ULONG baud, 
					   char parity, int numstopbits, int numdatabits, char flowctr)
{
	USHORT data;
	USHORT DLL;
	USHORT DLM;
	int devided_by_4 = 0;
	long status;

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;																//check port number
    }
	
    ioc9010->flowControl = flowctr;														//save flow control

    if(ioc9010->moduleType != SIGNATURE_8516)                                           // reset 8515 here since if it is 8516, the reset takes place in 8516 setup routine
    {
 	    data = 1 << portNo;      
	    status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, RESET_REG_85156, data);		//reset port
	}

	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//read LCR register
	//parity check
	data = data & PARITY_MASK_85156;												//clear parity bits
	switch (parity)
	{
	case 'N':
	case 'n':
		data = data | NO_PARITY_85156;												//set no parity
		break;
	case 'O':
	case 'o':
		data = data | ODD_PARITY_85156;												//set odd parity
		break;
	case 'E':
	case 'e':
		data = data | EVEN_PARITY_85156;											//set even parity
		break;
	default:
		status = PARITY_INVALID;												    
		return status;																//parity settings invalid
	}	
	
	//data bits
	data = data & DATABIT_MASK_85156;												//clear data bits settings
	switch (numdatabits)
	{
	case 5:
		data = data | DATA_5_85156;													//set 5 bits data
		break;
	case 6:
		data = data | DATA_6_85156;													//set 6 bits data
		break;
	case 7:
		data = data | DATA_7_85156;													//set 7 bits data
		break;
	case 8:
		data = data | DATA_8_85156;													//set 8 bits data
		break;
	default:
		status = DATA_BITS_INVALID;												    
		return status;																//data bit settings invalid
	}
	
	//stop bits
	data = data & STOPBIT_MASK_85156;												//clear stop bits settings
	switch (numstopbits)
	{
	case 1:
		data = data | STOPBIT_1_85156;												//set 1 stop bit
		break;
	case 2:
		data = data | STOPBIT_2_85156;												//set 2 or 1-1/2stop bits
		break;
	default:
		status = STOP_BITS_INVALID;												    
		return status;													//stop bit settings invalid
	}

	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//write to the memory

	data = data | ENABLE_DIVISOR_MASK_85156;										//enable divisor for baud rate settings
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//write to the memory

	//baud rate settings
	devided_by_4 = 0;
	switch (baud)
	{
	case 100:
		DLL = BAUDRATE_100_DLL_85156;												//set DLL
		DLM = BAUDRATE_100_DLM_85156;												//set DLM
		devided_by_4 = 1;
		break;
	case 400:
		DLL = BAUDRATE_400_DLL_85156;												//set DLL
		DLM = BAUDRATE_400_DLM_85156;												//set DLM
		break;
	case 600:
		DLL = BAUDRATE_600_DLL_85156;												//set DLL
		DLM = BAUDRATE_600_DLM_85156;												//set DLM
		devided_by_4 = 1;
		break;
	case 1200:
		DLL = BAUDRATE_1200_DLL_85156;												//set DLL
		DLM = BAUDRATE_1200_DLM_85156;												//set DLM
		devided_by_4 = 1;
		break;
	case 2400:
		DLL = BAUDRATE_2400_DLL_85156;												//set DLL
		DLM = BAUDRATE_2400_DLM_85156;												//set DLM
		break;
	case 4800:
		DLL = BAUDRATE_4800_DLL_85156;												//set DLL
		DLM = BAUDRATE_4800_DLM_85156;												//set DLM
		break;
	case 9600:
		DLL = BAUDRATE_9600_DLL_85156;												//set DLL
		DLM = BAUDRATE_9600_DLM_85156;												//set DLM
		break;
	case 19200:
		DLL = BAUDRATE_19200_DLL_85156;												//set DLL
		DLM = BAUDRATE_19200_DLM_85156;												//set DLM
		break;
	case 38400:
		DLL = BAUDRATE_38400_DLL_85156;												//set DLL
		DLM = BAUDRATE_38400_DLM_85156;												//set DLM
		break;
	case 57600:
		DLL = BAUDRATE_57600_DLL_85156;												//set DLL
		DLM = BAUDRATE_57600_DLM_85156;												//set DLM
		devided_by_4 = 1;
		break;
	case 76800:
		DLL = BAUDRATE_76800_DLL_85156;												//set DLL
		DLM = BAUDRATE_76800_DLM_85156;												//set DLM
		break;
	case 115200:
		DLL = BAUDRATE_115200_DLL_85156;											//set DLL
		DLM = BAUDRATE_115200_DLM_85156;											//set DLM
		devided_by_4 = 1;
		break;
	case 153600:
		DLL = BAUDRATE_153600_DLL_85156;											//set DLL
		DLM = BAUDRATE_153600_DLM_85156;											//set DLM
		break;
	case 230400:
		DLL = BAUDRATE_230400_DLL_85156;											//set DLL
		DLM = BAUDRATE_230400_DLM_85156;											//set DLM
		break;
	case 460800:
		DLL = BAUDRATE_460800_DLL_85156;											//set DLL
		DLM = BAUDRATE_460800_DLM_85156;											//set DLM
		break;
	case 921600:
		DLL = BAUDRATE_921600_DLL_85156;											//set DLL
		DLM = BAUDRATE_921600_DLM_85156;											//set DLM
		break;
	default:
		status = BAUD_RATE_INVALID;												    
		return status;																//baud rate settings error
	}
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + DLL_85156*2, 1, &DLL);//write to DLL
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + DLM_85156*2, 1, &DLM);//write to DLL

	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//read LCR register
	data = data & (~ENABLE_DIVISOR_MASK_85156);										//disable divisor to allow write
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LCR_85156*2, 1, &data);//write to LCR
	
	//clock divided by 4 settings
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//read MCR register
	if (devided_by_4)															    //prescaler needs to be devided by 4
		data = data | DIVIDED_BY_FOUR_85156;										//set MCR bit 7
	else
		data = data & (~DIVIDED_BY_FOUR_85156);										//clear MCR bit 7
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//write to MCR

	//hardware flow control
	if (flowctr & USE_HARDWARE_FLOW_TEST_85156)										//if need hardware flow control
	{
		status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + EFR_85156*2, 1, &data);//read EFR register
		data = data | RTS_DTR_CTS_DSR_ENABLE_85156;									//set enable
		status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + EFR_85156*2, 1, &data);//write to EFR
		
		status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//read MCR register
		if (flowctr & CTS_DTR_TEST_85156)	
			data = data | DTR_DSR_SELECT_85156;										//set DTR/DSR 
		else
			data = data & (~DTR_DSR_SELECT_85156);									//set CTS/RTS 
		status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &data);//write to MCR
	}

	//software flow control
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + EFR_85156*2, 1, &data);//read EFR register
	if (flowctr & SOFTWARE_TRANS_FLOW_85156)										//if need software flow control
		data = data | SOFTWARE_TRANS_FLOW_85156;										//set bot TX and RX Xon1 and Xon2 / Xoff1 and Xoff2
	else
		data = data & (~SOFTWARE_TRANS_FLOW_85156);									//clear software flow control
	if (flowctr & SOFTWARE_RCV_FLOW_85156)										//if need software flow control
		data = data | SOFTWARE_RCV_FLOW_85156;										//set bot TX and RX Xon1 and Xon2 / Xoff1 and Xoff2
	else
		data = data & (~SOFTWARE_RCV_FLOW_85156);									//clear software flow control
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + EFR_85156*2, 1, &data);//write to EFR
	
	//enable FIFO	
	data = ENABLE_FIFO_85156;
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCR_85156*2, 1, &data);//write to FCR

	//reset both TX RX FIFO
	data = data | FCR_RXFRESET_85156 | FCR_TXFRESET_85156;
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCR_85156*2, 1, &data);//write to FCR

//	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + ISR_85156*2, 1, &data);//read ISR register

	//clear all interrupts
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + IER_85156*2, 1, &data);//Read IER
	data =data & (~ENABLE_RHR_INT_85156) & (~ENABLE_THR_INT_85156) & (~ENABLE_RCVLINE_INT_85156)  & (~ENABLE_MODEM_INT_85156); 
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + IER_85156*2, 1, &data);//write to IER

	return status;																	//return status
}

/*----------------------------------------------------------------------------------*/
/*Function:		Configure 8516 COM Port for RS-485)     							*/
/*----------------------------------------------------------------------------------*/
/*Description:	set up one 8516 COM port											*/
/*Prototype:	Hy8516_ComsSetup(int port, USHORT baud, char parity,				*/
/*				int numstopbits, int numdatabits, char flowctr, int halfduplex,     */
/*              int delay485)                                						*/
/*Arguements:	same as above             											*/
/*				halfduplex = 0 Full duplex                                  		*/
/*					       = 1 Half duplex with echo                       			*/
/*					       = 2 Half duplex without echo                   			*/
/*				delay485 = A number between 0 and 15. This parameter is only used if*/
/*					halfduplexmode = 1. It specifies the delay time in terms of		*/
/*					bits (depends on baudrate) after the last stop bit of the last	*/
/*					transmitted character before the state of the RTS# output is	*/
/*					changed to enable receiving.									*/
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
long Hy8516_ComsSetup(struct ioc9010_uart_port *ioc9010, int portNo, ULONG baud, 
								 char parity, int numstopbits, int numdatabits, 
								 char flowctr, int halfduplex, int delay485)
{
	USHORT data;
	long status;

    if((delay485 < 0) || (delay485 > 15))
        return RS485_DELAYERROR;

    if((halfduplex < 0) || (halfduplex > 2))
        return RS485_DUPLEXERROR;

	//determine half/full duplex
    if(ioc9010->moduleType == SIGNATURE_8516)                                                    // only 8516 has half/full duplex settings. RS232 is always full duplex
    {
 	    data = 1 << portNo;      
	    status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, RESET_REG_85156, data);		//reset port

		ioc9010->duplex = halfduplex;
		ioc9010->withEcho = (halfduplex == 1)? 1 : 0;
		status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, &data);	    //read duplex setting reg
		if(halfduplex)                                                                  // only 8516 and half duplex need to set up something. Full duplex is the default */
		{
      		data = (~(1 << portNo)) & data;      
      		status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, data);	//set portNo duplex

			if(delay485 > 0)                                                            // if equals 0, don't need to do anything 
			{
				//enable auto RS485 half duplex
				status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCTR_85156*2, 1, &data);//read FCTR register
    			data = data | AUTO_RS485_ENABLE_85156;									//set auto RS485 enable to half duplex
				status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCTR_85156*2, 1, &data);//write to FCTR
	           
				//485 delay after transmitting
    			data = delay485 << 4;										            //set delay after 485 transmitting
				status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MSR_85156*2, 1, &data);//write to MSR
			}
		}else
		{
   		   data = (1 << portNo) | data;      
		   status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, data);	 //set port duplex
		}
	}

    /* call the Comsetup for the rest */
    return Hy85156_ComsSetup(ioc9010, portNo, baud, parity, numstopbits, numdatabits, flowctr);
}

/*----------------------------------------------------------------------------------*/
/*Function:		Enable Port Interrupt      										    */
/*----------------------------------------------------------------------------------*/
/*Description:	set up one or more interrupt enable bits for the specified port     */
/*Prototype:	Hy85156_EnableInterrupt(int port, int mask, int rcvTrigLevel,       */
/*              int transTrigLevel)                                                 */
/*Arguements:	portNo = serial channel 0 ~ 7										*/
/*				mask = interrupt enable mask                          	            */
/*                   = bit 7  6  5  4  3  2  1  0                                   */
/*                         |  |  |  |  |  |  |  |                                   */
/*                         |  |  |  |  |  |  | RX Data or RxFIFO up to a level INT  */
/*                         |  |  |  |  |  | TX Empty or TxFIFO down to a level INT  */
/*                         |  |  |  |  | RX Line Status INT                         */
/*                         |  |  |  | Modem Status INT                              */
/*                         |  |  | Not used                                         */
/*                         |  | Xon/Xoff Special Chat INT                           */
/*                         | RTS/DTR# INT                                           */
/*                        CTS/DSR# INT                                              */
/*				rcvTrigLevel = 1 ~ 64, interrupt trigger level on receiving			*/
/*                         only needed when mask bit0 and/or bit2 is set            */
/*				transTrigLevel = 1 ~ 64, interrupt trigger level on transmitting	*/
/*                         only needed when mask bit1 is set                        */
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
long Hy85156_EnableInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int mask, int rcvTrigLevel, int transTrigLevel)
{
	USHORT data;
	long status;

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;								//check port number
    }
		
	if ((mask & ENABLE_RHR_INT_85156) && ((rcvTrigLevel < 1) || (rcvTrigLevel > 64)))
   	{
		status = INT_TRIGGERLEVEL_INVALID;												    
		    return status;									//trigger level must be between 1 ~ 64
	}
		
	if ((mask & ENABLE_THR_INT_85156) && ((transTrigLevel < 1) || (transTrigLevel > 64)))
   	{
		status = INT_TRIGGERLEVEL_INVALID;												    
		    return status;								   //trigger level must be between 1 ~ 64
	}

	//clear all interrupt sources for all channels
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MSR_85156*2, 1, &data);//Read MSR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + ISR_85156*2, 1, &data);//Read ISR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//Read LSR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RHR_85156*2, 1, &data);//Read RHR
	
	//set up interrupt enable
	data = mask & 0xFF;
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + IER_85156*2, 1, &data);//write to IER

    if ((mask & ENABLE_RHR_INT_85156) || (mask & ENABLE_THR_INT_85156))
    {
    	//set up using level trigger
    	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCTR_85156*2, 1, &data);//Read FCTR
    	data =data | ENABLE_LEVEL_TRIGGER_85156; 
    	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCTR_85156*2, 1, &data);//write to FCTR
    }

    if (mask & ENABLE_RHR_INT_85156) 
    {	
    	data = (USHORT) rcvTrigLevel;
    	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RXTRG_85156*2, 1, &data);//write to RXTRG
    }

    if (mask & ENABLE_THR_INT_85156) 
    {
    	data = 64 - (USHORT) transTrigLevel;
    	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + TXTRG_85156*2, 1, &data);//write to TXTRG
    }

	//call enable interrupt
	status = EnableInterrupt(ioc9010, ioc9010->intLevel, 0);
	return status;																	//return status

}


/*----------------------------------------------------------------------------------*/
/*Function:		Disable Port Interrupt      										*/
/*----------------------------------------------------------------------------------*/
/*Description:	set up one or more interrupt enable bits for the specified port     */
/*Prototype:	Hy85156_DisableInterrupt(int portNo, int mask)                      */
/*Arguements:	portNo = serial channel 0 ~ 7										*/
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
long Hy85156_DisableInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int mask)
{
	long status = 0;
//	USHORT data;

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;								//check port number
    }

/*	//clear all interrupt sources for all channels
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MSR_85156*2, 1, &data);//Read MSR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + ISR_85156*2, 1, &data);//Read ISR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//Read LSR
	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RHR_85156*2, 1, &data);//Read RHR
	
	//disable relative interrupt 
	data = mask & 0xFF;
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + IER_85156*2, 1, &data);//write to IER
*/		
	//call disable interrupt
	status = DisableInterrupt(ioc9010, ioc9010->intLevel, 0);
	return status;																	//return status

}

/*----------------------------------------------------------------------------------*/
/*Function:		Get Interrupt Source												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Get which channel cause the interrtupt and what reason				*/
/*Prototype:	long Hy85156_INTReason(struct ioc9010_uart_port *ioc9010, 			*/
/*										int* portNo, int *reason)					*/
/*Arguements:	ioc9010 = uart port												    */
/*				*portNo = integer pointer to return the interrupt channel 0 ~ 7		*/
/*				*reason = integer pointer to the reason of the interrtupt           */
/*                      LSB (bit0 ~ bit7) major reason                              */
/*						= 0x00 MSR (Modem Status Register)							*/
/*						= 0x01 None													*/
/*						= 0x02 TXRDY (Transmitter Holding Register Empty)			*/
/*						= 0x04 RXRDY (received Data Ready)							*/
/*						= 0x06 LSR (Receiver Line Status Register)					*/
/*						= 0x0C RXRDY (Receive Data Timeout)							*/
/*						= 0x10 RXRDY (Received Xon/Xoff or Special character)		*/
/*						= 0x20 CTS#/DSR#, RTS#/DTR# change of state					*/
/*                      MSB (bit8 ~ bit15) detail reason                            */
/*                          when LSB = MSR                                          */
/*                              bit12 ~ bit15: not used                             */
/*                              bit11: CD# line interrupt                           */
/*                              bit10: RI# line interrupt                           */
/*                              bit9: DSR# line interrupt                           */
/*                              bit8: CTS# line interrupt                           */
/*                          when LSB = LSR                                          */
/*                              bit13 ~ bit15: not used                             */
/*                              bit12: RX Break                                     */
/*                              bit11: RX Framing Error                             */
/*                              bit10: RX Parity Error                              */
/*                              bit9: RX Overrun                                    */
/*                              bit8: not used                                      */
/*Return:		0 = OK																*/
/*				if error,  negative code in m_errorcode								*/
/*----------------------------------------------------------------------------------*/
long Hy85156_INTReason(struct ioc9010_uart_port *ioc9010, int* portNo, int *reason)
{
	long status;
	USHORT data = 0;
	USHORT detaildata = 0;
	int i;
	int found = 0;

	status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, INTERRUPT_REG_85156, &data);   //read interrupt back
	for (i=0; i<8; i++)
	{
		if(data & (1 << i))
		{
			*portNo = i;
			/* First read ISR to determine which interrupt or which area of interrupt */
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + ISR_85156*2, 1, &data);//Read ISR
            switch(data & 0x3F)
            {
                case INTERRUPT_REASON_MSR:						/* this is a MSR interrupt, get MSR */
			        status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + MSR_85156*2, 1, &detaildata);//Read ISR
                    break;
                case INTERRUPT_REASON_LSR:						/* this is a LSR interrupt, get LSR */
			        status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) i * PORT_OFFSET + LSR_85156*2, 1, &detaildata);//Read ISR
                    break;
            }
			*reason = (int) ((data & 0x3F) | ((detaildata & 0x0F) << 8));	
			found = 1;
			break;
		}
	}

	if (!found) 
        status = FALSE_INTERRUPT;			//if cannot find interrupt source, it is a false interrupt

	return status;
}

/*----------------------------------------------------------------------------------*/
/*Function:		Reset Interrupt														*/
/*----------------------------------------------------------------------------------*/
/*Description:	Reset interrtupt													*/
/*Prototype:	long Hy85156_ResetInterrupt(struct ioc9010_uart_port *ioc9010, 		*/
/*											int portNo, int reason)					*/
/*Arguements:	ioc9010 = uart port												    */
/*				portNo = interrupt channel 0 ~ 7									*/
/*				reason = the reason of the interrtupt 								*/
/*						= 0x00 MSR (Modem Status Register)							*/
/*						= 0x01 None													*/
/*						= 0x02 TXRDY (Transmitter Holding Register Empty)			*/
/*						= 0x04 RXRDY (received Data Ready)							*/
/*						= 0x06 LSR (Receiver Line Status Register)					*/
/*						= 0x0C RXRDY (Receive Data Timeout)							*/
/*						= 0x10 RXRDY (Received Xon/Xoff or Special character)		*/
/*						= 0x20 CTS#/DSR#, RTS#/DTR# change of state					*/
/*Return:		0 = OK																*/
/*				if error,  negative code in m_errorcode								*/
/*----------------------------------------------------------------------------------*/
long Hy85156_ResetInterrupt(struct ioc9010_uart_port *ioc9010, int portNo, int reason)
{
	long status=0;
	USHORT data;

	switch(reason & 0x3F)
	{
		case INTERRUPT_REASON_MSR:	//MSR interrupt
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MSR_85156*2, 1, &data);//Read MSR
			break;
		case INTERRUPT_REASON_NONE:	//None
			break;
		case INTERRUPT_REASON_TXRDY:	//TXRDY
		case INTERRUPT_REASON_RXRDYXONOFF:	//RXRDY (Received Xon/Xoff or Special character)
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + ISR_85156*2, 1, &data);//Read ISR
			break;
		case INTERRUPT_REASON_RXRDY:	//RXRDY  (received Data Ready)	
			break;
		case INTERRUPT_REASON_LSR:	//LSR (Receiver Line Status Register)
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//Read LSR
			break;
		case INTERRUPT_REASON_RXRDYTO:	//RXRDY (Receive Data Timeout)
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RHR_85156*2, 1, &data);//Read RHR
			break;
	}

	//re-enable interrupt for all -- 5331, 3331, carrier
	status = EnableInterrupt(ioc9010, ioc9010->intLevel, 0);

	return status;
}


/*----------------------------------------------------------------------------------*/
/*Function:		Write to general purpose line      									*/
/*----------------------------------------------------------------------------------*/
/*Description:	Write logic 1 or logic 0 to specified general purpose line			*/
/*Prototype:	long Hy85156_WriteLine(struct ioc9010_uart_port *ioc9010, 			*/
/*										int portNo, int line, int data)             */
/*Arguements:	ioc9010 = uart port												    */
/*				portNo = serial channel 0 ~ 7										*/
/*				line = 0 to use RTS                               					*/
/*                   = 1 to use DTR                                                 */              
/*                   It cannot use both at the same time                            */
/*				data = logic 1 or logic 0                 					        */
/*Return:		0 if successful      												*/
/*				if error, negative code in m_errorcode								*/
/*----------------------------------------------------------------------------------*/
long Hy85156_WriteLine(struct ioc9010_uart_port *ioc9010, int portNo, int line, int data)
{
	long status;
	USHORT udata;

	if (ioc9010->flowControl & USE_HARDWARE_FLOW_TEST_85156)																//if need hardware flow control
    {
        status = GENERAL_PURPOSE_INVALID;																			//when hardware control is selected, RTS/DTR cannot be used as general purpose line
    }else
    {
        status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &udata);      //Read MCR
        //set line level
        if (line)																									// use DTR. This doesn't work for XR16L788 chip as per its manual
        {
            if(data)																								// be careful here, logic level 0 and 1 are inverted in the register and the output
            	udata &= (~GENERAL_PURPOSE_DTR_85156); 
            else
            	udata |= GENERAL_PURPOSE_DTR_85156; 
            udata |= GENERAL_PURPOSE_USE_DTR_85156;
        }else																										// use RTS. Only this line can be used as general purpose line for XR16L788 chip
        {
            if(data)
            	udata &= (~GENERAL_PURPOSE_RTS_85156); 
            else
            	udata |= GENERAL_PURPOSE_RTS_85156; 
            udata &= (~GENERAL_PURPOSE_USE_DTR_85156);
        }

        status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + MCR_85156*2, 1, &udata);     //write to MCR
    }
	return status;
    
}

/*----------------------------------------------------------------------------------*/
/*Function:		Send Data															*/
/*----------------------------------------------------------------------------------*/
/*Description:	Write data to specified serial channel								*/
/*Prototype:	long Hy85156_Write(struct ioc9010_uart_port *ioc9010, int portNo, 	*/
/*									UCHAR *value, int length)						*/
/*Arguements:	ioc9010 = uart port												    */
/*				portNo = serial channel 0 ~ 7										*/
/*				*value = pointer to the 8 bit value array							*/
/*				length = length of the 8 bit value array							*/
/*Return:		number of bytes written												*/
/*				if error, negative code in m_errorcode								*/
/*----------------------------------------------------------------------------------*/
long Hy85156_Write(struct ioc9010_uart_port *ioc9010, int portNo, const UCHAR *value, int length)
{
	long status;
	USHORT data;
	int buffer_available;
	int number_to_send;
	int total_len;
	int i, counter;

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;																									//check port number
    }

	total_len = length;
	counter = 0;
    
    /*********************************************************************************/
    /* for half duplex, since the MAX1486 chip hard wire the TX to the RX so anything*/
    /* sent out would be echo back. When in full duplex mode, TX is not connected to */
    /* RX. Hence this switch over to avoid echoing.                                  */
    /*********************************************************************************/
    if((ioc9010->moduleType == SIGNATURE_8516) && (!ioc9010->withEcho))										// only 8516 has half/full duplex settings. RS232 is always full duplex
    {
        if(ioc9010->duplex)																						// only 8516 and half duplex need to set up something. Full duplex is the default */
        {
	        status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, &data);	//read duplex setting reg
	   	    data = (1 << portNo) | data;      
//		    status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, data);	//set port duplex
//            msleep(1);																									//delay 1ms then flush the receiving FIFO. This requirement is wield. Why?
        }
    }
	do
	{
		//read transmit FIFO counter
		status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + TXCNT_85156*2, 1, &data);//read TFCNT register
		buffer_available = FIFO_SIZE - (int) data;
		if (buffer_available > 0)
		{
			//setup the data length to send
			if (buffer_available >= total_len)
			{
				number_to_send = total_len;
				total_len = 0;
			}
			else
			{
				number_to_send = buffer_available;
				total_len -= number_to_send;
			}
			//loop pushing to the FIFO
			for(i=0; i<number_to_send;i++)
			{
				do
				{
					status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//read LSR register	
				}while(!(data & THR_EMPTY_85156));																		//check THR empty
				data = *(value + counter++);
				status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + THR_85156*2, 1, &data);//write to THR
			}
			status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + ISR_85156*2, 1, &data);//read ISR register	
			if(status < 0) break;
		}
	}while(total_len > 0);

    /*********************************************************************************/
    /* for half duplex, since the MAX1486 chip hard wire the TX to the RX so anything*/
    /* sent out would be echo back. When in full duplex mode, TX is not connected to */
    /* RX. Hence this switch back after the transmitting.                            */
    /* First we need to make sure the sending process is absolutely finished.        */
    /*********************************************************************************/
	do
	{
		status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + LSR_85156*2, 1, &data);//read LSR register	
	}while((!(data & THR_EMPTY_85156)) || (!(data & TSR_EMPTY_85156)));		    										//check THR empty

    if((ioc9010->moduleType == SIGNATURE_8516) && (!ioc9010->withEcho))										// only 8516 has half/full duplex settings. RS232 is always full duplex
    {
        if(ioc9010->duplex)																						// only 8516 and half duplex need to set up something. Full duplex is the default */
        {
	        status = IOC9010IPRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, &data);	//read duplex setting reg
	      	data = (~(1 << portNo)) & data;      
//	      	status = IOC9010IPWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, DUPLEX_REG_85156, data);	//set port duplex        
//            msleep(1);																									//delay 1ms then flush the receiving FIFO. This requirement is wield. Why?
//            Hy85156_Flush_Single(ioc9010, portNo, 0, 1);
        }
    }
	
	if (status == 0) status = length;																					//successfully send length bytes
	return status;
}

/*----------------------------------------------------------------------------------*/
/*Function:		Read receving values												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read from receiving buffer of one channel in the 8515 IP card		*/
/*Prototype:	long Hy85156_Read(struct ioc9010_uart_port *ioc9010, int portNo, 	*/
/*									UCHAR *value)									*/
/*Arguements:	ioc9010 = uart port												    */
/*				portNo = serial channel 0 ~ 7										*/
/*				*value = pointer to the 8 bit value array							*/
/*Return:		number of bytes read												*/
/*				if error,  negative code in m_errorcode								*/
/*----------------------------------------------------------------------------------*/
long Hy85156_Read(struct ioc9010_uart_port *ioc9010, int portNo, UCHAR *value)
{
	long status;
	USHORT data;
	USHORT RXCounter;
	int i, realCount;

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;								//check port number
    }
	
	//read receiver FIFO counter
    realCount = 0;
    do
    {
	   status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RXCNT_85156*2, 1, &RXCounter);//read RFCNT register
	   if (RXCounter > 0)
	   {	
		  for(i=0; i<RXCounter; i++)
		  {
              status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + RHR_85156*2, 1, &data);//read data from buffer to value pointer
		      if (status == 0)
              {
				 if((UCHAR) data != 0)		/* This is to kill the false character 0 receiveing due to the half/full duplex switching */
				 {
                 	*(value + realCount) = (UCHAR) data;
                 	realCount++;
				 }
              }
			  else
			     break;
		  }
	   }
    }while((RXCounter > 0) && (status == 0));

	return (status>=0) ? (long) realCount : status;
}

/*----------------------------------------------------------------------------------*/
/*Function:		Flush the serial buffer												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Flush a serial buffer of the specified channel						*/
/*Prototype:	long Hy85156_Flush(struct ioc9010_uart_port *ioc9010, int portNo, 	*/
/*									int TX, int RX)									*/
/*Arguements:	ioc9010 = uart port												    */
/*				portNo = serial channel 0 ~ 7										*/
/*				TX = true to flush the sending buffer, false not					*/
/*				RX = true to flush the receiving buffer, false not					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
long Hy85156_Flush_Single(struct ioc9010_uart_port *ioc9010, int portNo, int TX, int RX)
{
	long status;
	USHORT data;
	
	if ((!TX) && (!RX)) return 0;													//if neither need TX flush nor RX, don't do anything

	if (portNo < 0 || portNo >7) 
    {
		status = PORT_INVALID_85156;												    
        return status;								//check port number
    }

	status = IOC9010IPMemoryRead(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCR_85156*2, 1, &data);//read FCR register
	if (TX) data = data | FLUSH_TX_FIFO;											//request TX FIFO flushing
	if (RX) data = data | FLUSH_RX_FIFO;											//request RX FIFO flushing
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCR_85156*2, 1, &data);//write to FCR
	data = data | ENABLE_FIFO_85156;											    //enable FIFO
	status = IOC9010IPMemoryWrite(ioc9010->virtureMemBaseAddr, ioc9010->ipSlot, (ULONG) portNo * PORT_OFFSET + FCR_85156*2, 1, &data);//write to FCR

	return status;
}

/*----------------------------------------------------------------------------------*/
/*Function:		Flush multi-serial buffers											*/
/*----------------------------------------------------------------------------------*/
/*Description:	Flush serial buffers according to the mask							*/
/*Prototype:	long Hy85156_Flush(struct ioc9010_uart_port *ioc9010, USHORT mask)	*/
/*Arguements:	ioc9010 = uart port												    */
/*				mask = serial channel mask											*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
long Hy85156_Flush(struct ioc9010_uart_port *ioc9010, USHORT mask)
{
	long status;
	int i;
	int flush_TX, flush_RX;

	for (i=0; i<8; i++)
	{
		flush_TX = (mask & (1 << (i*2)))?1:0;								//need TX flush?
		flush_RX = (mask & (1 << (i*2+1)))?1:0;								//need RX flush?
		status = Hy85156_Flush_Single(ioc9010, i, flush_TX, flush_RX);
		if (!status) break;															//if error, no need to stay longer
	}
	return status;
}


/*----------------------------------------------------------------------------------------------*/
/*Function:		Setup interrupt (overloading for concurrent device driver						*/
/*----------------------------------------------------------------------------------------------*/
/*Description:	Setup interrupt for 5331, 3331, 8002 or IOC9010	or Tip120						*/
/*Prototype:	long SetupInterrupt(struct ioc9010_uart_port * ioc9010, int level, int side)	*/
/*Arguements:	ioc9010 = uart port															    */
/*				level = Not used																*/
/*				side = 0, A side, = 1, B side													*/
/*Return:		error code and also in m_ErrorStatus											*/
/*				if successful, return 0															*/
/*----------------------------------------------------------------------------------------------*/
long EnableInterrupt(struct ioc9010_uart_port *ioc9010, int level, int side)
{
	unsigned short sdata, offset;
	int status = 0;

	switch (ioc9010->device_ID)
	{
		case 0x9010:								//device is IOC9010
		case 0x6335:								//device is 6335
		case 0x7002:								//device is 7002
			//set up interrupt on PLX chip
			if(ioc9010->device_ID == 0x7002)
			{
				sdata = inw(ioc9010->carrierAddr + HYTEC7002CSR);
				sdata |= 0x0001;
				outw(sdata, ioc9010->carrierAddr + HYTEC7002CSR);		//enable IRQ in FPGA. added by Jim on 13-10-2010
			}
			else
			{
				offset = (ioc9010->device_ID == 0x9010) ? PLXLOCALCONFIGCSR : ((ioc9010->device_ID == 0x7002) ? HYTEC7002CSR : PLXLOCALCONFIGCSR8311);
				sdata = inw(ioc9010->plxLocalConfigBAR + offset);
				sdata |= (ioc9010->device_ID == 0x9010) ? 0x043 : ((ioc9010->device_ID == 0x7002) ? 0x0001 : 0x0900);
				outw(sdata, ioc9010->plxLocalConfigBAR + offset);		//enable IRQ on PLX. added by Jim on 26-03-2009
			}

       		/* Enable Carrier ALL Board Interrupts */
	        sdata = inw(ioc9010->carrierAddr + REG_IP_INT_MASK);
			sdata |= 1 << ioc9010->ipSlot * 2;						//shift to the right bit. Note: Hytec IP only uses INT0
       		outw(0x0000, ioc9010->carrierAddr + REG_IP_ERROR_MASK);	//01/12/2010 JC, this causes grief when IP card not pull it high. So set it to 0!
   			outw(sdata, ioc9010->carrierAddr + REG_IP_INT_MASK); 		//enable the IP interrupt
			break;
		default:
			status = WRONG_DEVICE;											//wrong device
			break;
	}
	return status;
}

/*----------------------------------------------------------------------------------------------*/
/*Function:		Disable interrupt 						                                        */
/*----------------------------------------------------------------------------------------------*/
/*Description:	Setup interrupt for 5331, 3331, 8002 or IOC9010	or Tip120						*/
/*Prototype:	long DisableInterrupt(struct ioc9010_uart_port *ioc9010, int level, int side)   */
/*Arguements:	ioc9010 = uart port															    */
/*				level = Not used																*/
/*				side = 0, A side, = 1, B side													*/
/*Return:		error code and also in m_ErrorStatus											*/
/*				if successful, return 0															*/
/*----------------------------------------------------------------------------------------------*/
long DisableInterrupt(struct ioc9010_uart_port *ioc9010, int level, int side)
{
	unsigned short sdata, offset;
	int status = 0;

	switch (ioc9010->device_ID)
	{
		case 0x9010:								//note: when device is IOC9010, vmeslot = 0
		case 0x6335:								//note: when device is 6335, vmeslot = 0
		case 0x7002:								//note: when device is 7002, vmeslot = 0
			if(ioc9010->device_ID == 0x7002)
			{
				sdata = inw(ioc9010->carrierAddr + HYTEC7002CSR);
				sdata &= (~0x0001);
				outw(sdata, ioc9010->carrierAddr + HYTEC7002CSR);		//clear IRQ in FPGA. added by Jim on 13-10-2010
			}
			else
			{
				offset = (ioc9010->device_ID == 0x9010) ? PLXLOCALCONFIGCSR : ((ioc9010->device_ID == 0x7002) ? HYTEC7002CSR :PLXLOCALCONFIGCSR8311);
				sdata = inw(ioc9010->plxLocalConfigBAR + offset);
				sdata &= (ioc9010->device_ID == 0x9010) ? (~0x040) : ((ioc9010->device_ID == 0x7002) ? (~0x0001) : (~0x0900));
				outw(sdata, ioc9010->plxLocalConfigBAR + offset);		//clear IRQ on PLX. added by Jim on 26-03-2009
			}
			break;
		default:
			status = WRONG_DEVICE;		//wrong device
			break;
	}

	return status;
}

